package mscJNeuralNet.connectors;

import mscJNeuralNet.net.Net;

/**
 * Schnittstelle fr alle Klassen, welche die Kantengewichte eines KNN initialisieren.
 * Es existieren verschiedene Algorithmen, um den Initialwert eines 
 * Kantengewichts so zu bestimmen, dass weder ein Lernverfahren in einem lokalen 
 * Minima hngenbleibt, noch die Dimensionalitt der Gewichtsmatrix nicht voll ausgeschpft wird.
 * Beispiel fr ein bekanntes Problem, das die Leistungsfhigkeit eines KNN einschrnkt ist
 * das Symmetrie-Problem, bei dem bestimmte Kantengewichtswerte ein lernen verhindern.
 * Entsprechend gibt es "Symmetrybreaking" Algorithmen, mit denen solche Gewichtskonstellationen
 * verhindert werden sollen.<p/>
 * 
 * In der aktuellen Version dieser API werden lediglich Feedforward-Netze
 * ohne Shortcut-Verbindungen untersttzt. Daher sollte der Connector auch nur die
 * jeweiligen Gewichte initialisieren.<br/>
 * Am einfachsten werden die BIAS und die restlichen Gewichte getrennt initialisiert.
 * Zuerst werden die BIAS-Werte initalisiert:
 * <pre>
 * // first connect BIAS to all but inputlayer
 * for (int lInputtingLayer = 1;
 * 		lInputtingLayer<lLayerNeurons.length;
 * 		lInputtingLayer++){
 * 	for (lInputtingNeuronIndex=0;
 * 		lInputtingNeuronIndex<lLayerNeurons[lInputtingLayer].length;
 * 		lInputtingNeuronIndex++){
 * 		if (lInputtingLayer==1)
 * 			lWeights[0][lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex]] =
 * 				Net.getRandomWeightValue();
 *		else
 *			lWeights[0][lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex]] =
 *				Net.getRandomWeightValue());
 *	}
 * }
 * </pre>
 * Hierbei zu beachten sind die Indizes, mit denen die Neuronen und das BIAS in der Gewichte-
 * Matrix reprsentiert und auch adressiert werden. Das BIAS hat den Index 0, gefolgt
 * von den Eingabeneuronen, den verdeckten Neuronen und den Ausgabeneuronen.<br/>
 * Anschlieend knnen die Schichten schichtweise miteinander verbunden werden.
 * Empfohlen wird eine systematische Schleife von der Eingabeschicht hin zur Ausgabeschicht.
 * <pre>
 * for (lOutputtingLayer=0; lOutputtingLayer<lLayerNeurons.length-1; lOutputtingLayer++){
 * 	lInputtingLayer = lOutputtingLayer+1;
 * 	for (lOutputtingNeuronIndex=0; 
 * 			lOutputtingNeuronIndex<lLayerNeurons[lOutputtingLayer].length;
 * 			lOutputtingNeuronIndex++){
 * 		for (lInputtingNeuronIndex=0;
 * 			lInputtingNeuronIndex<lLayerNeurons[lInputtingLayer].length;
 * 			lInputtingNeuronIndex++){
 * 			if (lInputtingLayer==1)
 * 				lWeights[lLayerNeurons[lOutputtingLayer][lOutputtingNeuronIndex]]
 * 						[lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex]] =
 * 							Net.getRandomWeightValue();
 * 			else
 * 				lWeights[lLayerNeurons[lOutputtingLayer][lOutputtingNeuronIndex]]
 * 						[lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex]] =
 * 							Net.getRandomWeightValue();
 * 		}
 * 	}
 * </pre>
 * <p/>
 * 
 * Created on 02.06.2004
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 02.06.2004
 */
public interface INetConnector {
	
	/**
	 * Initialisiert die Gewichtsmatrix des gegebenen Netzes mit Startwerten,
	 * so da das Netz verbunden (connected) ist.<br/>
	 * blicherweise besitzt ein bergebenes Netz eine Gewichtsmatrix,
	 * die nur mit 0-Werten gefllt ist und daher keine Verbindungen erahnen lt.
	 * Nach der Initialisierung sollten die Observer des Netzes benachrichtigt werden.
	 * 
	 * @param pNet Das Netz, dessen Kantengewichte initialisiert werden sollen.
	 * @param pParameters Je nach Initialisierungsalgorithmus knnen hier verschiedene Parameter
	 * 			bermittelt werden.
	 */
	public void connectNet(Net pNet, double[] pParameters);
}
