package mscJNeuralNet.netPerformanceStatistics;


/**
 * Diese Klasse enthlt Informationen zu verschiedenen Netzfehlern.
 * Die verfgbaren Netzfehlertypen werden in dieser Klasse festgelegt.
 * <p/>
 * Da diese Klasse sehr viele Informationen ber den aktuellen Fehlerzustand
 * des Netzes speichert, gibt es auch eine "Leichte"-Version dieser Klasse, die
 * Klasse {@link mscJNeuralNet.netPerformanceStatistics.LightWeightNetPerformanceStatistics}, die lediglich
 * die Durchschnittswerte sowie den max. neuron ABSE speichert.
 * <p/>
 * Eine Instanz mit berechneten Werten wird durch den Aufruf der Methode 
 * {@link mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatisticsCalculator#calculateErrors(Net, double[][], double[][])}
 * erzeugt:<br>/
 * <pre>
 * // Netz erzeugen und initialisieren
 * int [] lLayerSizes = {2, 1};
 * Net myNet = new Net(lLayerSizes);
 * new RandomSymmetryBreakingNetConnector().connectNet(myNet, null);
 * 
 * // Lerndaten fr Xor 
 * double [][] lXorInput = {{1,1}, {1,-1}, {-1, 1}, {-1,-1}};
 * double [][] lXorOutput = {{-1}, {1}, {1}, {-1}};
 * 
 * // Aktuelle Netzperformaz berechnen
 * NetPerformanceStatistics lActualStats = 
 * 	NetPerformanceStatisticsCalculator.calculateErrors(
 * 		myNet, lXorInput, lXorOutput);
 *  
 * System.out.println(STR_averageSumAbsoluteError_short+"="+
 * 	Double.toString(lActualStats.averageSumOfAbsoluteErrors));
 * </pre>
 * <p/>
 * Created on 06.06.2004
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 06.06.2004
 * @see mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatisticsCalculator
 * @see mscJNeuralNet.netPerformanceStatistics.LightWeightNetPerformanceStatistics
 * @see mscJNeuralNet.netPerformanceStatistics
 */
public class NetPerformanceStatistics {
	
	/** Kurze Stringreprsentation des Fehlertyps avg. ABSE. */
	public static final String STR_averageSumAbsoluteError_short = "avg. ABSE";
	/** Kurze Stringreprsentation des Fehlertyps avg. SSE. */
	public static final String STR_averageSumOfSquaredError_short = "avg. SSE";
	/** Kurze Stringreprsentation des Fehlertyps avg. RMSE. */
	public static final String STR_averageRootMeanSquareError_short = "avg. RMSE";
	/** Kurze Stringreprsentation des Fehlertyps max. neuron ABSE. */
	public static final String STR_maximumNeuronAbsoluteError_short = "max. neuron ABSE";

	/** Konstante fr den Fehlertyp avg. ABSE. */
	public static final int ERRORTYPE_averageSumAbsoluteError = 0;
	/** Konstante fr den Fehlertyp avg. SSE. */
	public static final int ERRORTYPE_averageSumOfSquaredError = 1;
	/** Konstante fr den Fehlertyp avg. RMSE. */
	public static final int ERRORTYPE_averageRootMeanSquareError = 2;
	/** Konstante fr den Fehlertyp max. neuron ABSE. */
	public static final int ERRORTYPE_maximumNeuronAbsoluteError = 3;
	
	/** Lange Stringreprsentation des Fehlertyps avg. ABSE. */
	public static final String STR_averageSumAbsoluteError_long = "average sum of absolute errors";
	/** Lange Stringreprsentation des Fehlertyps avg. SSE. */
	public static final String STR_averageSumOfSquaredError_long = "average sum of squared errors";
	/** Lange Stringreprsentation des Fehlertyps avg. RMSE. */
	public static final String STR_averageRootMeanSquareError_long = "average root mean square errors";
	/** Lange Stringreprsentation des Fehlertyps max. neuron ABSE. */
	public static final String STR_maximumNeuronAbsoluteError_long = "maximum ABSE by a single output neuron";
	
	/** 
	 * Array aller kurzen Stringreprsentation der Fehlertypen.
	 * Reihenfolge:<br/>
	 * <dir>
	 * 	<li>avg. ABSE</li>
	 *  <li>avg. SSE</li>
	 * 	<li>avg. RMSE</li>
	 * 	<li>max. neuron ABSE</li>
	 * </dir> 
	 */
	public static final String [] STR_errorTypes_short = {
		STR_averageSumAbsoluteError_short,
		STR_averageSumOfSquaredError_short,
		STR_averageRootMeanSquareError_short,
		STR_maximumNeuronAbsoluteError_short
	};
	
	/**
	 * [pattern][neuron]</br>
	 * Die Eingabemsuter, die dem Netz prsentiert werden.
	 */
	public double [][] inputPatterns;

	/**
	 * [pattern][neuron]</br>
	 * Die Ausgabemuster, die das Netz lernen soll (Soll-Werte).
	 */
	public double [][] outputPatterns;
		
	/**
	 * [pattern][neuron]</br>
	 * Die Netzausgaben zu den entsprechenden Eingabemustern (Ist-Werte).
	 * The output of the net to the given inputpattern
	 */
	public double netOutputs[][];
		
	// neuron errors
		
	/**
	 * [pattern][neuron]</br>
	 * Absoluter Fehler eines einzelnen Neurons, berechnet als Abs(Soll-Wert - Ist-Wert).
	 */
	public double neuronAbsoluteError[][];

	/**
	 * [pattern][neuron]</br>
	 * Quadrat des Fehlers eines einzelnen Neurons, berechnet als
	 * (Soll-Wert - Ist-Wert)^2
	 */
	public double neuronSquaredError[][];
	
	/**
	 * Der grte absolute Fehlerwert eines einzelnen Neurons,
	 * berechnet als Abs(Soll-Wert - Ist-Wert).
	 */
	public double maximumNeuronAbsoluteError;
	
	/**
	 * Index des Ausgabeneurons mit dem grten
	 * absoluten Fehlerwert.
	 */
	public int maximumNeuronAbsoluteErrorNeuronIndex;
	
	/**
	 * Index des Lerndatensatzes mit dem grten absoluten Fehlerwert fr
	 * ein Neuron.
	 */
	public int maximumNeuronAbsoluteErrorPatternIndex;
		
	// pattern errors
		
	/**
	 * [pattern]</br>
	 * Absoluter Fehler eines einzelnen Lerndatensatzes, berechnet als
	 * Summe fr alle Ausgabeneuronen ( Abs(Soll-Wert - Ist-Wert) )
	 */
	public double [] sumOfAbsoluteErrors;

	/**
	 * [pattern]</br>
	 * Quadrierter Fehler eines einzelnen Lerndatensatzes, berechnet als
	 * Summe fr alle Ausgabeneuronen ( (Soll-Wert - Ist-Wert)^2 )
	 */
	public double [] sumOfSquaredErrors;

	/**
	 * [pattern]</br>
	 * Quadratwurzel des quadriereten Fehlers eines einzelnen Lerndatensatzes,
	 * berechnet als 
	 * Sqrt(Quadrierter Fehler eines einzelnen Lerndatensatzes / Anzahl Ausgabeneuronen)
	 */
	public double [] rootMeanSquareErrors;
		
	// average all patterns errors
		
	/**
	 * Durchschnitt der Summe aller absoluten Fehlerbetrge (avg. ABSE)
	 */
	public double averageSumOfAbsoluteErrors;

	/**
	 * Durchschnitt der Fehlerquadratsummen aller Lerndatenstze (avg. SSE)
	 */
	public double averageSumOfSquaredErrors;
		
	/**
	 * Durchschnitt der Quadratwurzeln der quadratischen
	 * Fehler aller Lerndatenstze (avg. RMSE) 
	 */
	public double averageRootMeanSquareError;
	
	/**
	 * Liefert den Fehlerwert, der durch den bergebenen Parameter (ERRORTYPE)
	 * festgelegt wurde.
	 * 
	 * @param pErrorType Eine Konstante ERRORTYPE fr den Fehlertypen.
	 * @return Wert des Fehlers.
	 */
	public double getErrorByType(int pErrorType){
		switch (pErrorType){
			case ERRORTYPE_averageSumAbsoluteError:
				return averageSumOfAbsoluteErrors;
			case ERRORTYPE_averageSumOfSquaredError:
				return averageSumOfSquaredErrors;
			case ERRORTYPE_averageRootMeanSquareError:
				return averageRootMeanSquareError;
			case ERRORTYPE_maximumNeuronAbsoluteError:
				return maximumNeuronAbsoluteError;
		}
		return 0;
	}
}
