package mscJNeuralNet.patterns;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.Writer;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.ResourceBundle;
import java.util.Vector;

import logger.Log;
import formats.Formats;

/**
 * Diese Klasse reprsentiert Lerndatenstze fr das KNN.
 * Die Lerndatenstze werden als Eingabemuster-Array und Ausgabemuster-Array
 * gespeichert.<br/>
 * Folgende Features sind mglich:<br/>
 * <dir>
 * 	<li>Die Lerndatenstze in einem 
 * 		<a href="http://www-ra.informatik.uni-tuebingen.de/SNNS/">SNNS</a>
 * 		-kompatiblen Format gespeichert (@link #save(String, Patterns)) 
 * 		und geladen (@link #load(String))werden.
 *	</li>
 *	<li>Es ist mglich SNNS-hnliche Kommentarzeilen mit dem Kommentarsymbol "#"
 * 		in die gespeicherten Netze einzufgen.
 *	</li>
 *	<li>Die Menge der Lerndatenstze kann nach widersprchlichen
 * 		Datenstzen durchsucht werden ({@link #getContradictoryPatterns()}).
 *	</li>
 *	<li>Mehrere Lerndatensatz-Mengen (Patterns Objekte) knnen zu einem einzigen
 *		Lerndatensatz zusammengefgt werden ({@link #merge(Patterns, Patterns)}).
 *	</li>
 *	<li>Die Lerndatenstze knne in einer zuflligen Reihenfolge
 *		abgerufen werden ({@link #getRandomizedPatternsOrder()}).
 *	</li>
 * </dir>
 * <p/>
 * Created on 10.11.2003
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 14.06.2004
 */
public class Patterns {
	private static final ResourceBundle smMSCResourceBundle = ResourceBundle.getBundle("mscJNeuronalNet_Text");
	
	// TODO: methode zum entfernen von doppelten eingabe/ausgabemuster paaren
	
	private static final ResourceBundle smResourceBundle = ResourceBundle.getBundle("debug");
	private static final boolean smLogMethodCall;
	static{
		if (smResourceBundle.getString("showMethodCall.Patterns").equalsIgnoreCase("true"))
			smLogMethodCall = true;
		else smLogMethodCall = false;
	}
	// Mi Jun 4 04:44:59 2003
	private static SimpleDateFormat smSNNSSimpleDateFormat = new SimpleDateFormat("EE MMM dd HH:mm:ss yyyy");
	
	/** Array, in dem das Eingabemuster aufbewahrt werden soll. patternindex / neuronindex. */
	private double[][] mInputPatterns = null;
	/** Array, in dem das Ausgabemuster aufbewahrt werden soll. patternindex / neuronindex. */
	private double[][] mOutputPatterns = null;
	
	/**
	 * Field inputsize
	 */
	private int mInputPatternsSize;

	/**
	 * Field outputsize
	 */
	private int mOutputPatternsSize;
	
	private Vector mComments = new Vector();
	
	public static void main(String argv[]){
		Patterns pats = null;
		if (argv.length>0){
			try {
				pats = Patterns.load(argv[0]);
				Log.getInstance().writeln(pats.toString());
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
		else Log.getInstance().writeln("Use parameter to load a patternfile.");
	}
	
	/** 
	 * Erzeugt eine neue Instanz von Patterns.
	 * Die gegebenen Parameter werden benutzt, um
	 * den bentigten Speicher fr die Lerndaten zu reservieren.
	 *  
	 * @param inputsize Gre eines Eingabemusters.
	 * @param outputsize Gre eines Ausgabemusters.
	 * @param patterncount Anzahl der zu speichernden Lerndatenstze.
	 */
	public Patterns(int inputsize, int outputsize, int patterncount){
		mInputPatterns = new double[patterncount][inputsize];
		mOutputPatterns = new double[patterncount][outputsize];
		this.mInputPatternsSize = inputsize;
		this.mOutputPatternsSize = outputsize;
	}
	
	
	/**
	 * Liefert die Eingabemuster aller Lerndatenstze.
	 * 
	 * @return Eingabemuster aller Lerndatenstze.
	 */
	public double[][] getInputPatterns(){   return mInputPatterns;}

	/**
	 * Liefert das Eingabemuster des Lerndatensatzes mit dem gegebenen Index.
	 * 
	 * @param index Index des Lerndatensatzes.
	 * @return Eingabemuster des Lerndatensatzes mit dem gegebenen Index.
	 */
	public double[] getInputPattern(int index){   return mInputPatterns[index];}

	/**
	 * Liefert die Ausgabemuster aller Lerndatenstze.
	 * 
	 * @return Ausgabemuster aller Lerndatenstze.
	 */
	public double[][] getOutputPatterns(){   return mOutputPatterns;}

	/**
	 * Liefert das Ausgabemuster des Lerndatensatzes mit dem gegebenen Index.
	 * 
	 * @param index Index des Lerndatensatzes.
	 * @return Ausgabemuster des Lerndatensatzes mit dem gegebenen Index.
	 */
	public double[] getOutputPattern(int index){   return mOutputPatterns[index];}

	/**
	 * Liefert ein einzelnes Datum des angegebenen Eingabemusters.
	 * 
	 * @param neuronindex Index des Neurons, zu dem das Datum bestimmt werden soll.
	 * @param patternindex Index des Lerndatensatzes, aus dem das Datum bestimmt werden soll.
	 * @return Einzelnes Datum des Eingabemusters.
	 */
	public double getInputToken(int neuronindex, int patternindex){
		return mInputPatterns[patternindex][neuronindex];
	}

	/**
	 * Liefert ein einzelnes Datum des angegebenen Ausgabemusters.
	 * 
	 * @param neuronindex Index des Neurons, zu dem das Datum bestimmt werden soll.
	 * @param patternindex Index des Lerndatensatzes, aus dem das Datum bestimmt werden soll.
	 * @return Einzelnes Datum des Ausgabemusters.
	 */
	public double getOutputToken(int neuronindex, int patternindex){
		return mOutputPatterns[patternindex][neuronindex];
	}

	/**
	 * Liefert die Gre eines Eingabemusters.
	 * 
	 * @return Die Gre eines Eingabemusters.
	 */
	public int getInputPatternSize(){ return mInputPatternsSize;}
	
	/**
	 * Liefert die Gre eines Ausgabemusters.
	 * 
	 * @return Die Gre eines Ausgabemusters.
	 */
	public int getOutputPatternSize(){ return mOutputPatternsSize;}

	/**
	 * Liefert die Anzahl der Lerndatenstze.
	 * 
	 * @return Die Anzahl der Lerndatenstze.
	 */
	public int getPatternsCount(){  return mInputPatterns.length;}
//		  public void setInputPatterns(double [][] inputpatterns){    inpat = inputpatterns;}
//		  public void setOutputPatterns(double [][] outputpatterns){    outpat = outputpatterns;}

	/**
	 * Legt den Wert eines einzelnen Datums im definierten Eingabemuster fest.
	 * 
	 * @param neuronindex Index des Neurons, zu dem das Datum gendert werden soll.
	 * @param patternindex Index des Lerndatensatzes, zu dem das Datum gendert werden soll.
	 * @param value Der neue Wert fr das Datum.
	 */
	public void setInputToken(int neuronindex, int patternindex, double value){ 
		mInputPatterns[patternindex][neuronindex] = value;
	}

	/**
	 * Legt den Wert eines einzelnen Datums im definierten Ausgabemuster fest.
	 * 
	 * @param neuronindex Index des Neurons, zu dem das Datum gendert werden soll.
	 * @param patternindex Index des Lerndatensatzes, zu dem das Datum gendert werden soll.
	 * @param value Der neue Wert fr das Datum.
	 */
	public void setOutputToken(int neuronindex, int patternindex, double value){
		mOutputPatterns[patternindex][neuronindex] = value;
	}
	
	/**
	 * Liefert ein Array mit allen Indizes der Lerndatenstze in zuflliger Reihenfolge.
	 * 
	 * @return Array mit allen Indizes der Lerndatenstze in zuflliger Reihenfolge.
	 */
	public int[] getRandomizedPatternsOrder(){
		int lPatternsCount = getPatternsCount();
		int[] lSelectionList = new int[lPatternsCount];
		int i,j;
		int[] lUsed = new int[lPatternsCount];
		
		// initialize the local working list
		for (i=0; i<lPatternsCount; i++)
			lUsed[i] = 0;

		// generate the random order selection list

		for (i=0; i<lPatternsCount; i++)  {
			
			j= ((int)(Math.random()*lPatternsCount)) % lPatternsCount;
			while (lUsed[j] == 1)
				j = (j+1) % lPatternsCount;
			lUsed[j] = 1;
			lSelectionList[i] = j;
		}
		
		return lSelectionList;
	}
	
	/**
	 * Liefert einen Vector aller Kommentare fr diese Lerndatenmenge.
	 * Die einzelnen Eintrge des Vectors sind vom Typ String.
	 * Die Kommentare werden beim Speichern der Lerndatenmenge als Kommentarzeilen
	 * eingefgt.
	 * 
	 * @return Vector aller Kommentare fr diese Lerndatenmenge.
	 */
	public Vector getComments(){
		return mComments;
	}
	
	/**
	 * Entfernt alle Kommentare fr diese Lerndatenmenge.
	 * Die Kommentare werden beim Speichern der Lerndatenmenge als Kommentarzeilen
	 * eingefgt.
	 */
	public void removeComments(){
		mComments.removeAllElements();
	}
	
	/**
	 * Fgt eine neue Kommentarzeile zu dieser Lerndatenmenge.
	 * Die Kommentare werden beim Speichern der Lerndatenmenge als Kommentarzeilen
	 * eingefgt. 
	 * 
	 * @param pComment Neue Kommentarzeile zu dieser Lerndatenmenge.
	 */
	public void addComment(String pComment){
		mComments.add(pComment);
	}
	
	/**
	 * berprft, ob gleiche Eingabemuster verschiedenen Ausgabemsutern zugeordnet sind.
	 * Zwei Muster sind widersprchlich, wenn sie dasselbe Eingabemuster, aber verschiedene
	 * Ausgabemuster haben.
	 * 
	 * @return Zwei Lerndatenindizes, falls die Lerndaten einen Widerspruch enthalten, sonst null.
	 */
	public int [] getContradictoryPatterns(){
		int lSourcePatternIndex, lTargetPatternIndex;
		int [] lContradictionaryPatterns = new int[2];
		for (lSourcePatternIndex = 0; lSourcePatternIndex<getPatternsCount(); lSourcePatternIndex++){
			for (lTargetPatternIndex = lSourcePatternIndex+1; lTargetPatternIndex<getPatternsCount(); lTargetPatternIndex++){
				// test if two input patterns are equal
				if (
					arePatternsEqual(mInputPatterns[lSourcePatternIndex], mInputPatterns[lTargetPatternIndex])
					&&
					!arePatternsEqual(mOutputPatterns[lSourcePatternIndex], mOutputPatterns[lTargetPatternIndex])
				){
					lContradictionaryPatterns[0] = lSourcePatternIndex;
					lContradictionaryPatterns[1] = lTargetPatternIndex;
					return lContradictionaryPatterns;
				}
			}
		}
		return null;
	}
	
	/**
	 * Testet, ob zwei Muster (Arrays mit double-Werten) gleich sind.
	 * 
	 * @param pPatternOne Erstes Muster.
	 * @param pPatternTwo Zweites Muster.
	 * @return true, falls beide Muster gleich sind, false, wenn sie verschieden sind.
	 */
	public static boolean arePatternsEqual(double [] pPatternOne, double [] pPatternTwo){
		for (int lNeuronIndex = 0; lNeuronIndex<pPatternOne.length; lNeuronIndex++){
			if (pPatternOne[lNeuronIndex] != pPatternTwo[lNeuronIndex])
				return false;
		}
		return true;
	}
	
	public String toString(){
		StringBuffer s = new StringBuffer(200);
		s.append("Patterns in:"+getInputPatternSize()+" out:"+getOutputPatternSize()+" pat:"+getPatternsCount()+"\n");
		for (int i=0; i<mInputPatterns.length; i++){
			s.append("pat:"+Formats.format(i+"", 3, 3)+"  ");
			s.append(Formats.format(mInputPatterns[i]));
			s.append("=> ");
			s.append(Formats.format(mOutputPatterns[i])+"\n");
		}
		return s.toString();
	}
	
	/**
	 * Ldt alle Lerndatenstze (Eingabe- sowie Ausgabemuster) aus einer Datei
	 * und erzeugt ein Patterns Objekt.
	 * 
	 * @param pSourceFileName Dateiname der Lerndatenstze.
	 * @return Lerndatenstze aus der angegebenen Datei. 
	 * @throws IOException Falls beim Laden der Lerndatenstze aus der Datei ein Fehler entsteht.
	 * @see #load(Reader)
	 */
	public static Patterns load(String pSourceFileName) throws IOException{
		Log.getInstance().writeln("Loading patternfile "+pSourceFileName+"...", true);
		FileReader theFile = new FileReader(pSourceFileName);
		Patterns pat = load(theFile);
		Log.getInstance().writeln(pSourceFileName+" in:"+pat.getInputPatternSize()+" out:"+pat.getOutputPatternSize()+" pat:"+pat.getPatternsCount());
		theFile.close();
		return pat;
	}
	
	/**
	 * Ldt alle Lerndatenstze (Eingabe- sowie Ausgabemuster) aus einer Datei
	 * und erzeugt ein Patterns Objekt.
	 * 
	 * @param pSourceFile Datei der Lerndatenstze.
	 * @return Lerndatenstze aus der angegebenen Datei. 
	 * @throws IOException Falls beim Laden der Lerndatenstze aus der Datei ein Fehler entsteht.
	 * @see #load(Reader)
	 */
	public static Patterns load(File pSourceFile) throws IOException{
		Log.getInstance().writeln("Loading patternfile "+pSourceFile+"...", true);
		FileReader theFile = new FileReader(pSourceFile);
		Patterns pat = load(theFile);
		Log.getInstance().writeln(pSourceFile.getName()+" in:"+pat.getInputPatternSize()+" out:"+pat.getOutputPatternSize()+" pat:"+pat.getPatternsCount());
		theFile.close();
		return pat;
	}
	
	/**
	 * Ldt alle Trainingsdaten (Eingabe- sowie Ausgabemuster) aus einem
	 * Characterstream (Reader) und erzeugt ein Patternobjekt.
	 * Die gelesene Datei sollte das von SNNS benutzte Dateiformat fr Patterns sein:
	 * <pre>
	 * SNNS pattern definition file V3.2
	 * generated at Wed Feb 04 14:34:21 2004
	 * 
	 * # Ein Kommentar
	 * 
	 * No. of patterns : 147
	 * No. of input units : 4
	 * No. of output units : 1
	 * 
	 * 0.0 1.0 1.0 0.5 -0.3
	 * 1.0 0.0 -1.0 0.2 0.1
	 * ...
	 * </pre>
	 * Dabei sind die Zeilen "SNNS pattern definition file V3.2" und
	 * "generated at Wed Feb 04 14:34:21 2004" optional.<br/>
	 * Die Anzahl der Lerndatenstze (No. of patterns), Eingabegre (No. of input units) und 
	 * Ausgabegre (No. of output units) knnen auch ohne das
	 * vorangestellte "No. of ... :" angegeben werden, wenn die folgende Reihenfolge eingehalten
	 * wird:<br/> 
	 * Anzahl Lerndatenstze<br/>
	 * Eingabegre<br/>
	 * Ausgabegre<br/>
	 * Bsp.: <br/>
	 * <pre>
	 * 147 4 1
	 * </pre>
	 * Zwischen den eigentlichen Datenstzen und dem Header mit
	 * Anzahl der Datenstze, Eingabegre und Ausgabegre MUSS mindestens 
	 * ein Zeilenvorschub stehen:
	 * <pre>
	 * 2 1 1
	 * 1.0 1.0 1.0 1.0 0.0 0.0
	 * </pre>
	 * Ein Fehler wre:
	 * <pre>
	 * 2 1 1 1.0 1.0 1.0
	 * 1.0 0.0 0.0
	 * </pre>
	 * Die Daten selbst knnen beliebig durch Zeilenvorschbe und Leerzeichen formatiert 
	 * werden.
	 * Der bersichtlichkeit halber sollte in jeder Zeile ein Datensatz stehen und
	 * zwischen den Eingabe- und Ausgabemustern ein sichtbarer Leerzeichenabstand sein.
	 * 
	 * @param pReader Reader, aus dem die Lerndatenstze gelesen werden sollen.
	 * @return Lerndatenstze aus dem angegebenen Reader. 
	 * @throws IOException Falls beim Laden der Lerndatenstze aus der Datei ein Fehler entsteht.
	 */
	public static Patterns load(Reader pReader) throws IOException{
		BufferedReader lBIn = new BufferedReader(pReader);
		String lLine;
		int lCommentIndex;
		Vector lComments = new Vector();
		int lInputSize = 0;
		int lOutputSize = 0;
		int lPatternsAmount = 0;
		int lTmp;
		String [] lSplitLine;
				
		while (lInputSize==0 || lOutputSize==0 || lPatternsAmount==0){
			lLine = lBIn.readLine().trim();
			// auf kommentar prfen
			if ((lCommentIndex = lLine.indexOf("#")) > -1){
				lComments.add(lLine.substring(lCommentIndex, lLine.length()));
				lLine = lLine.substring(0, lCommentIndex).trim();
			}
			
			if (lLine.length()>0){
				// auf zeile "SNNS pattern definition file V3.2" prfen
				if (lLine.indexOf("SNNS pattern definition")<0){
					// auf zeile "generated at Mi Jun 4 04:44:59 2003" prfen
					if (lLine.indexOf("generated at")<0){
						// No. of patterns : 147
						// No. of input units : 4 
						// No. of output units : 1
						if (lLine.indexOf("No. of patterns :")>-1 && lPatternsAmount==0){
							lTmp = lLine.indexOf(":");
							lPatternsAmount = Integer.parseInt(lLine.substring(lTmp+1, lLine.length()).trim());
						}
						else if (lLine.indexOf("No. of input units :")>-1 && lInputSize==0){
							lTmp = lLine.indexOf(":");
							lInputSize = Integer.parseInt(lLine.substring(lTmp+1, lLine.length()).trim());
						}
						else if (lLine.indexOf("No. of output units :")>-1  && lOutputSize==0){
							lTmp = lLine.indexOf(":");
							lOutputSize = Integer.parseInt(lLine.substring(lTmp+1, lLine.length()).trim());
						}
						else{
							// kein SNNS format. dennoch versuchen, ob das nchste token vielleicht
							// ein integer ist.
							lSplitLine = lLine.split("\\s+");
							for (int i=0; i<lSplitLine.length; i++){
								if (lPatternsAmount==0) lPatternsAmount = Integer.parseInt(lSplitLine[i]);
								else if (lInputSize==0) lInputSize = Integer.parseInt(lSplitLine[i]);
								else if (lOutputSize==0) lOutputSize = Integer.parseInt(lSplitLine[i]);
							}
						}
					}
				}
			}
		}
		
		Patterns lLoadedPattern = new Patterns(lInputSize, lOutputSize, lPatternsAmount);
		
		// kommentare bertragen
		for (int i=0; i<lComments.size(); i++)
			lLoadedPattern.addComment((String) lComments.get(i));
		
		// now, read the data records
		int lActualNeuronIndex=0;
		int lActualPattern=0;
		int lTokensToLoad = (lInputSize+lOutputSize)*lPatternsAmount;
		int lTokensLoaded = 0;
		boolean lInputPattern = true;
		while (lTokensLoaded<lTokensToLoad){
			lLine = lBIn.readLine();
			if ((lCommentIndex = lLine.indexOf("#")) > -1)
				lLine = lLine.substring(0, lCommentIndex).trim();
			if (lLine.length()>0){
				lSplitLine = lLine.split("\\s+");
				for (int i=0; i<lSplitLine.length; i++){
					if (lInputPattern){
						lLoadedPattern.setInputToken(lActualNeuronIndex++, lActualPattern, Double.parseDouble(lSplitLine[i]));
						if (lActualNeuronIndex >= lInputSize){
							lActualNeuronIndex = 0;
							lInputPattern = false;
						}
					}
					else{
						lLoadedPattern.setOutputToken(lActualNeuronIndex++, lActualPattern, Double.parseDouble(lSplitLine[i]));
						if (lActualNeuronIndex >= lOutputSize){
							lActualNeuronIndex = 0;
							lInputPattern = true;
							lActualPattern++;
						}
					}
					lTokensLoaded++;
				}
			}
		}
		return lLoadedPattern;
	}
	
	/**
	 * Speichert die gegebene Lerndatenmenge in der gegebenen Datei.
	 * 
	 * @param pTargetFileName Dateiname, in welche die Lerndatenmenge gespeichert werden soll.
	 * @param pPatternsToSave Zu speichernde Lerndatenmenge.
	 * @throws IOException Falls bei der Ausgabe der Lerndatenstze in die Datei ein
	 * 						Fehler entsteht.
	 * @see #load(Reader)
	 */
	public static void save(String pTargetFileName, Patterns pPatternsToSave) throws IOException{
		if (smLogMethodCall) Log.getInstance().writeln("Saving patternfile "+pTargetFileName+"...");
		FileWriter theFile = new FileWriter(pTargetFileName);
		save(theFile, pPatternsToSave);
	}
	
	/**
	 * Speichert die gegebene Lerndatenmenge in der gegebenen Datei.
	 * 
	 * @param pTargetFile Datei, in welche die Lerndatenmenge gespeichert werden soll.
	 * @param pPatternsToSave Zu speichernde Lerndatenmenge.
	 * @throws IOException Falls bei der Ausgabe der Lerndatenstze in die Datei ein
	 * 						Fehler entsteht.
	 * @see #load(Reader)
	 */
	public static void save(File pTargetFile, Patterns pPatternsToSave) throws IOException{
		if (smLogMethodCall) Log.getInstance().writeln("Saving patternfile "+pTargetFile.getName()+"...");
		FileWriter theFile = new FileWriter(pTargetFile);
		save(theFile, pPatternsToSave);
	}
	
	/**
	 * Gibt die gegebene Lerndatenmenge ber das gegebene Writer-Objekt aus.
	 * Das Benutze Dateiformat wird unter {@link #load(Reader)} beschrieben.
	 * 
	 * @param pWriter Writer, in welche die Ausgabe stattfinden soll.
	 * @param pPatternToSave Zu speichernde Lerndatenmenge.
	 * @see #load(Reader)
	 */
	public static void save(Writer pWriter, Patterns pPatternToSave){
		if (smLogMethodCall) Log.getInstance().writeln("Saving patternfile.");
		PrintWriter out = new PrintWriter(pWriter);
		
		/*
			SNNS pattern definition file V3.2
			generated at Mi Jun 4 04:44:59 2003
		*/
		out.println("SNNS pattern definition file V3.2");
		out.println("generated at "+smSNNSSimpleDateFormat.format(Calendar.getInstance().getTime()));
		
		Vector lComments = pPatternToSave.getComments();
		for (int i=0; i<lComments.size(); i++)
			out.println("# "+((String)lComments.get(i)));
		
		out.println();
		
		/*
			No. of patterns : 54
			No. of input units : 15
			No. of output units : 3
		*/
		out.println("No. of patterns : "+Integer.toString(pPatternToSave.getPatternsCount()));
		out.println("No. of input units : "+Integer.toString(pPatternToSave.getInputPatternSize()));
		out.println("No. of output units : "+Integer.toString(pPatternToSave.getOutputPatternSize()));
		
		out.println();
		
		int i, j;
		// now, write the data records
		for (i=0; i<pPatternToSave.getPatternsCount(); i++) {
			for (j=0; j<pPatternToSave.getInputPatternSize(); j++)   // read the input values part of the pattern
				out.print(Double.toString(pPatternToSave.getInputToken(j, i))+" ");
			out.print("   ");
			for (j=0; j<pPatternToSave.getOutputPatternSize(); j++)  // then read the expected values part
				out.print(Double.toString(pPatternToSave.getOutputToken(j, i))+" ");
			out.println();
		}
		out.flush();
		out.close();
	}
	
	/**
	 * Erzeugt eine neue Lerndatenmenge aus den beiden bergebenen Lerndatenmengen.
	 * Es ist notwendig, dass beide Patterns Objekte identische Eingabe- und 
	 * Ausgabegrssen haben.
	 * 
	 * @param pPatternOne Erste Lerndatenmenge.
	 * @param pPatternTwo Zweite Lerndatenmenge.
	 * @return Eine neue Lerndatenmenge, das aus den beiden bergebenen Lerndatenmengen besteht.
	 * @throws PatternsIncompatibleException Falls beide Lerndatenmengen nicht identische 
	 * 											Eingabe- und Ausgabegrssen haben.
	 */
	public static Patterns merge(Patterns pPatternOne, Patterns pPatternTwo) throws PatternsIncompatibleException{
		if (smLogMethodCall) {
			Log.getInstance().writeln("Patterns.merge(...)"); 
			Log.getInstance().writeln(" one: "+pPatternOne.getInputPatternSize()+"+"+pPatternOne.getOutputPatternSize()+" x "+pPatternOne.getPatternsCount());
			Log.getInstance().writeln(" two: "+pPatternTwo.getInputPatternSize()+"+"+pPatternTwo.getOutputPatternSize()+" x "+pPatternTwo.getPatternsCount());
		}
		if (pPatternOne.getInputPatternSize() != pPatternTwo.getInputPatternSize()) throw
			new PatternsIncompatibleException(smMSCResourceBundle.getString("INCOMPATIBLE_INPUT_PATTERNSIZES"));
		if (pPatternOne.getOutputPatternSize() != pPatternTwo.getOutputPatternSize()) throw
			new PatternsIncompatibleException(smMSCResourceBundle.getString("INCOMPATIBLE_OUTPUT_PATTERNSIZES"));
		Patterns lMergedPatttern = new Patterns(pPatternOne.getInputPatternSize(), 
			pPatternOne.getOutputPatternSize(), pPatternOne.getPatternsCount()+pPatternTwo.getPatternsCount());
		
		int pat, neuron;
		// copy pattern one
		for (pat = 0; pat<pPatternOne.getPatternsCount(); pat++){
			// copy inputpattern
			for (neuron = 0; neuron<pPatternOne.getInputPatternSize(); neuron++)
				lMergedPatttern.setInputToken(neuron, pat, pPatternOne.getInputToken(neuron, pat));
			// copy outputpattern
			for (neuron = 0; neuron<pPatternOne.getOutputPatternSize(); neuron++)
				lMergedPatttern.setOutputToken(neuron, pat, pPatternOne.getOutputToken(neuron, pat));
		}
		// copy pattern two
		for (pat = 0; pat<pPatternTwo.getPatternsCount(); pat++){
			// copy inputpattern
			for (neuron = 0; neuron<pPatternTwo.getInputPatternSize(); neuron++)
				lMergedPatttern.setInputToken(neuron, pat+pPatternOne.getPatternsCount(), pPatternTwo.getInputToken(neuron, pat));
			// copy outputpattern
			for (neuron = 0; neuron<pPatternTwo.getOutputPatternSize(); neuron++)
				lMergedPatttern.setOutputToken(neuron, pat+pPatternOne.getPatternsCount(), pPatternTwo.getOutputToken(neuron, pat));
		}
		
		// kommentare bertragen
		Vector lComments = pPatternOne.getComments();
		for (int i=0; i<lComments.size(); i++)
			lMergedPatttern.addComment((String)lComments.get(i));
		lComments = pPatternTwo.getComments();
		for (int i=0; i<lComments.size(); i++)
			lMergedPatttern.addComment((String)lComments.get(i));
		
		return lMergedPatttern;
	}
	
	/* teilt das gegebene Pattern in zwei disjunkte Patterns. 1.pattern: 0 bis (exklusiv) pSplitIndex
	 * 2. pat: pSplitIndex (inklusive) bis rest */
	public static Patterns[] split(Patterns pPattern, int pSplitIndex) {
		Patterns[] lPatRet = new Patterns[2];
		lPatRet[0] = new Patterns(pPattern.getInputPatternSize(), 
				pPattern.getOutputPatternSize(), pSplitIndex);
		lPatRet[1] = new Patterns(pPattern.getInputPatternSize(), 
				pPattern.getOutputPatternSize(), pPattern.getPatternsCount()-pSplitIndex);
		for (int pat=0; pat<pSplitIndex; pat++) {
			for (int intoken=0; intoken<pPattern.getInputPatternSize(); intoken++)
				lPatRet[0].setInputToken(intoken, pat, pPattern.getInputToken(intoken, pat));
			for (int outtoken=0; outtoken<pPattern.getOutputPatternSize(); outtoken++)
				lPatRet[0].setOutputToken(outtoken, pat, pPattern.getOutputToken(outtoken, pat));
		}
		for (int pat=pSplitIndex; pat<pPattern.getPatternsCount(); pat++) {
			for (int intoken=0; intoken<pPattern.getInputPatternSize(); intoken++)
				lPatRet[1].setInputToken(intoken, pat-pSplitIndex, pPattern.getInputToken(intoken, pat));
			for (int outtoken=0; outtoken<pPattern.getOutputPatternSize(); outtoken++)
				lPatRet[1].setOutputToken(outtoken, pat-pSplitIndex, pPattern.getOutputToken(outtoken, pat));
		}
		
		return lPatRet;
	}
	
	public static Patterns[] splitRandomly(Patterns pPattern, int pAmountFirstPattern){
		int [] lRandPatterns = pPattern.getRandomizedPatternsOrder();
		Patterns[] lPatRet = new Patterns[2];
		lPatRet[0] = new Patterns(pPattern.getInputPatternSize(), 
				pPattern.getOutputPatternSize(), pAmountFirstPattern);
		lPatRet[1] = new Patterns(pPattern.getInputPatternSize(), 
				pPattern.getOutputPatternSize(), pPattern.getPatternsCount()-pAmountFirstPattern);
		for (int pat=0; pat<pAmountFirstPattern; pat++) {
			for (int intoken=0; intoken<pPattern.getInputPatternSize(); intoken++)
				lPatRet[0].setInputToken(intoken, pat, pPattern.getInputToken(intoken, lRandPatterns[pat]));
			for (int outtoken=0; outtoken<pPattern.getOutputPatternSize(); outtoken++)
				lPatRet[0].setOutputToken(outtoken, pat, pPattern.getOutputToken(outtoken, lRandPatterns[pat]));
		}
		for (int pat=pAmountFirstPattern; pat<pPattern.getPatternsCount(); pat++) {
			for (int intoken=0; intoken<pPattern.getInputPatternSize(); intoken++)
				lPatRet[1].setInputToken(intoken, pat-pAmountFirstPattern, pPattern.getInputToken(intoken, lRandPatterns[pat]));
			for (int outtoken=0; outtoken<pPattern.getOutputPatternSize(); outtoken++)
				lPatRet[1].setOutputToken(outtoken, pat-pAmountFirstPattern, pPattern.getOutputToken(outtoken, lRandPatterns[pat]));
		}
		
		return lPatRet;
	}
}
