package mscJNeuralNet.trainer;

import java.util.ResourceBundle;

import logger.Log;
import mscJNeuralNet.net.Net;
import mscJNeuralNet.net.PatternDoesNotMatchNetException;
import mscJNeuralNet.netPerformanceStatistics.NetPerformanceReporter;
import mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatistics;
import mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatisticsCalculator;
import mscJNeuralNet.patterns.Patterns;
import mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm;
import observerPattern.Observable;
import observerPattern.ObserverManager;

/**
 * Diese Klasse dient zum Trainieren eines KNN.
 * Mit Hilfe dieser Klasse kann das Trainieren des Netzes
 * als nebenlufiger Prozess realisiert werden.
 * Gleichzeitig stehen nicht-nebenlufige Methoden zur Verfgung,
 * um ein Netz bis zu einem bestimmten Fehlerwert oder einem bestimmten
 * Lernzyklus zu trainieren.
 * <p/>
 * <h3>Nicht-Nebenlufiges Trainieren</h3>
 * ber die statischen Methoden dieser Klasse kann einfach Trainiert werden:<br/>
 * <dir>
 * 	<li>{@link #train(INetTrainingAlgorithm, double[][], double[][], int)}:
 * 		Trainieren bis eine bestimmte Anzahl an Lernschritten absolviert wurde.
 * 	</li>
 *  <li>{@link #train(INetTrainingAlgorithm, double[][], double[][], double, int)}:
 * 		Trainieren bis zu einem bestimmten Fehlerwert. Es werden alle Haupt-Fehlertypen
 * 		aus {@link mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatistics} 
 * 		untersttzt.
 * 	</li>
 * 	<li>{@link #train(INetTrainingAlgorithm, double[][], double[][], double, int, int)}:
 * 		Trainieren bis zu einem bestimmten Fehlerwert oder bis eine bestimmte
 * 		Anzahl an Lernschritten absolviert wurde.
 * 	</li>
 * </dir>
 * <p/>
 * <h3>Nebenlufiges Trainieren</h3>
 * <p/>
 * Das nebenlufige Trainieren erlaubt es, das Training in einem eigenen Prozess 
 * ablaufen zu lassen. Das Training kann entweder durch Aufruf der Methode {@link #stop()}
 * oder durch erreichen bestimmter Bedingungen, wie dem Unterschreiten eines Fehlerwertes
 * oder das Absolvieren einer bestimmten Anzahl an Lernschritten angehalten werden.
 * Das Training kann durch die Schnittstelle {@link observerPattern.Observable}
 * berwacht werden, denn der NetTrainer benachrichtigt nach jedem Lernschritt alle gemeldeten
 * Observer. Im folgenden ein einfaches Beispiel, dass das dreidimensionale Xor-Problem als
 * nebenlufigen Prozess trainiert:<br/>
 * <pre>
package mscJNeuralNet.examples;

import observerPattern.Observable;
import observerPattern.Observer;
import mscJNeuralNet.connectors.INetConnector;
import mscJNeuralNet.connectors.RandomSymmetryBreakingNetConnector;
import mscJNeuralNet.net.Net;
import mscJNeuralNet.net.PatternDoesNotMatchNetException;
import mscJNeuralNet.netPerformanceStatistics.NetPerformanceReporter;
import mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatistics;
import mscJNeuralNet.patterns.Patterns;
import mscJNeuralNet.trainer.NetTrainer;
import mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm;
import mscJNeuralNet.trainingAlgorithms.RProp;

public class TestTrippleXorLearnConcurrent implements Observer{
	
  public TestTrippleXorLearnConcurrent() 
    throws PatternDoesNotMatchNetException{
    
    // 1. Erzeugen der bentigten Klassen
		
    // MLP {@link mscJNeuralNet.net.Net}
    // MLP mit der gewnschten Schichtstruktur erstellen:
    // Eingabeschicht: 3 Neuronen
    // 1. Verdeckte Schicht: 3 Neuronen
    // Ausgabeschicht: 1 Neuron
    int [] layerSizesTrippleXOr = {3, 3, 1};
    // {@link mscJNeuralNet.net.Net} myNet = new Net(layerSizesTrippleXOr);
    Net myNet = new Net(layerSizesTrippleXOr);

    // BIAS wurde automatisch bercksichtigt.

    // {@link mscJNeuralNet.connectors.INetConnector}
    // Diese Klasse wird zum Initialisieren der Kantengewichte bentigt
    INetConnector lNetConnectionAlgo = new RandomSymmetryBreakingNetConnector();
		
    // {@link mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm}
    // Diese Klasse wird zum Trainieren eines MLP bentigt
    INetTrainingAlgorithm lNetTrainAlgo = new RProp();
		
    // {@link mscJNeuralNet.trainer.NetTrainer}
    // Diese Klasse wird das Netz als nebenlufigen Prozess trainieren.
    NetTrainer lTrainer = new NetTrainer();
		
		
    // 2. Initialisieren des Netzes
		
    // MLP + {@link mscJNeuralNet.connectors.INetConnector}
    // Mit der Instanz von {@link mscJNeuralNet.connectors.INetConnector} inititalisieren.
    // Die Klasse {@link mscJNeuralNet.connectors.RandomSymmetryBreakingNetConnector} bentigt
    // keine eigenen Kontrollparameter und wird daher mit dem Wert null aufgerufen.
    lNetConnectionAlgo.connectNet(myNet, null);
		
    // Nun ist das Netz verbunden und initialisiert.
		
    // {@link mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm}
    // Lernverfahren {@link mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm} mit Netz verbinden.
    lNetTrainAlgo.setNet(myNet);
		
		
    // 3. Trainer konfigurieren
		
    // Dem Trainer das Lernverfahren und das MLP mitteilen 
    lTrainer.setTrainingAlgorithm(lNetTrainAlgo);
    // 100 Lernschritte lang trainieren
    lTrainer.setTargetCycles(100);
    // oder bis Fehlerwert unter 0.01
    lTrainer.setTargetError(0.01);
    // Fehlerwert soll vom Typ SSE sein
    lTrainer.setTargetErrorType(
      NetPerformanceStatistics.ERRORTYPE_averageSumOfSquaredError);
    // Lerndaten immer in derselben Reihenfolge prsentieren
    lTrainer.setUseRandomizedPatternOrder(false);
		
    // Dieses Programm als Observer des Trainers anmelden
    lTrainer.getObserverManager().addObserver(this);
		
		
    // 4. Trainieren des Netzes
		
    // TRAININGSDATEN
    // Trainingsdaten bereitstellen:
    // Trainingsdaten in ein Patterns Objekt bertragen
    // Komforbaler ist es, die Lerndaten in einer Textdatei 
    // zu speichern und hier zu laden
    Patterns lTrippleXorPat = new Patterns(3, 1, 8);
    // {-1,-1,-1}
    lTrippleXorPat.setInputToken(0, 0, -1);
    lTrippleXorPat.setInputToken(1, 0, -1);
    lTrippleXorPat.setInputToken(2, 0, -1);
    // {-1}
    lTrippleXorPat.setOutputToken(0, 0, -1);
		
    // {-1,-1,1}
    lTrippleXorPat.setInputToken(0, 1, -1);
    lTrippleXorPat.setInputToken(1, 1, -1);
    lTrippleXorPat.setInputToken(2, 1, 1);
    // {1}
    lTrippleXorPat.setOutputToken(0, 1, 1);

    // {-1,1,-1}
    lTrippleXorPat.setInputToken(0, 2, -1);
    lTrippleXorPat.setInputToken(1, 2, 1);
    lTrippleXorPat.setInputToken(2, 2, -1);
    // {1}
    lTrippleXorPat.setOutputToken(0, 2, 1);

    // {-1,1,1}
    lTrippleXorPat.setInputToken(0, 3, -1);
    lTrippleXorPat.setInputToken(1, 3, 1);
    lTrippleXorPat.setInputToken(2, 3, 1);
    // {-1}
    lTrippleXorPat.setOutputToken(0, 3, -1);

    // {1,-1,-1}
    lTrippleXorPat.setInputToken(0, 4, 1);
    lTrippleXorPat.setInputToken(1, 4, -1);
    lTrippleXorPat.setInputToken(2, 4, -1);
    // {1}
    lTrippleXorPat.setOutputToken(0, 4, 1);
		
    // {1,-1,1}
    lTrippleXorPat.setInputToken(0, 5, 1);
    lTrippleXorPat.setInputToken(1, 5, -1);
    lTrippleXorPat.setInputToken(2, 5, 1);
    // {-1}
    lTrippleXorPat.setOutputToken(0, 5, -1);

    // {1,1,-1}
    lTrippleXorPat.setInputToken(0, 6, 1);
    lTrippleXorPat.setInputToken(1, 6, 1);
    lTrippleXorPat.setInputToken(2, 6, -1);
    // {-1}
    lTrippleXorPat.setOutputToken(0, 6, -1);

    // {1,1,1}
    lTrippleXorPat.setInputToken(0, 7, 1);
    lTrippleXorPat.setInputToken(1, 7, 1);
    lTrippleXorPat.setInputToken(2, 7, 1);
    // {1}
    lTrippleXorPat.setOutputToken(0, 7, 1);
		
    // Trainingsdatenmenge dem Trainer mitteilen
    lTrainer.setTrainingPatterns(lTrippleXorPat);

    // Training beginnen
    lTrainer.start();
  }

  public static void main(String [] args) {
    try{
      new TestTrippleXorLearnConcurrent();
    }
    catch (PatternDoesNotMatchNetException e){
      e.printStackTrace();
    }
  }

  public void notify(Observable pObservable) {
    // Diese Methode wird vom NetTrainer nach jedem 
    // Lernschritt aufgerufen.
    // Das bergebene Observable Objekt ist der NetTrainer selbst
    if (pObservable instanceof NetTrainer){
      NetTrainer lTrainer = (NetTrainer) pObservable;
      // Prfen, ob das Training vorbei ist
      if (lTrainer.hasFinished()){
        System.out.println("Training fertig.");
        System.out.println(
          NetPerformanceReporter.getNetPerformance(
            lTrainer.getLastCalcualtedNetStatistics(), 
            lTrainer.getTrainingAlgorithm().getCycle()
          )
        );
        System.exit(0);
      }
      else{
        // Sonst alle 10 Lernschritte Status ausgeben
        if (lTrainer.getTrainingAlgorithm().getCycle() % 10 == 0){
          System.out.println("Lernschritt: "+
            lTrainer.getTrainingAlgorithm().getCycle());
          System.out.println("Aktueller Netzfehler: "+
            lTrainer.getTargetErrorTypeString()+" "+
            lTrainer.getLastCalculatedError());
        }
      }
    }
  }
}
 * </pre>
 * <p/>
 * Created on 05.06.2004
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 26.06.2004
 * @see Net
 * 
 */
public class NetTrainer implements Runnable, Observable{
	private static final ResourceBundle smMSCResourceBundle = ResourceBundle.getBundle("mscJNeuronalNet_Text");
	private static final ResourceBundle smResourceBundle = ResourceBundle.getBundle("debug");
	private static final boolean smLogMethodCall;
	private static boolean smLogTrainState = true;
	private static int smLogTrainStateIntervall;
	private static boolean smFinalReport = true;
	static{
		if (smResourceBundle.getString("showMethodCall.NetTrainer").equalsIgnoreCase("true"))
			smLogMethodCall = true;
		else smLogMethodCall = false;
		if (smResourceBundle.getString("showTrainingState.NetTrainer").equalsIgnoreCase("true"))
			smLogTrainState = true;
		else smLogTrainState = false;
		smLogTrainStateIntervall = Integer.parseInt(smResourceBundle.getString("showTrainingState.NetTrainer.Intervall"));
	}

	private Thread mThread = null;
	private boolean mRunning = false;
	private final ObserverManager mObserverManager = new ObserverManager();
	
	private double mTargetError = 0.01;
	private int mTargetErrorType = NetPerformanceStatistics.ERRORTYPE_maximumNeuronAbsoluteError;
	private int mTargetCycles = 1000;
	private boolean mUseRandomizedPatternOrder = true;
	
	private INetTrainingAlgorithm mTrainingAlgorithm;
	private Patterns mTrainingPatterns;
	private Patterns mValidationPatterns;
	private boolean mCheckTargetErrorOnValidationPattern = false;
	private NetPerformanceStatistics mLastCalculatedTrainNetStatistics;
	private NetPerformanceStatistics mLastCalculatedValidationNetStatistics;
	private double mLastCalculatedTrainError = 0D;
	private double mLastCalculatedValidationError = 0D;
	private double mLastCalculatedError = 0D;
	
	// TODO
	private double mRandomNoise;
	
	public void setTargetErrorOnValidationPatterns(boolean pCheckTargetErrorOnValidationPatterns) {
		mCheckTargetErrorOnValidationPattern = pCheckTargetErrorOnValidationPatterns;
	}

	
	private double mMinTrainError;
	private int mMinTrainErrorCycle;
	private double mMinValidationError;
	private int mMinValidationErrorCycle;
	
	
	public double getMinError() {
		return mMinTrainError;
	}
	
	public int getMinErrorCycle() {
		return mMinTrainErrorCycle;
	}
	
	private int mUpdateIntervall = 100;
	
	public void setUpdateIntervall(int pUpdateIntervallInCycles) {
		mUpdateIntervall = Math.max(1, pUpdateIntervallInCycles);
	}
	
	public int getUpdateIntervall() {
		return mUpdateIntervall;
	}
	
	/**
	 * Legt das Lernverfahren fr das Training fest.
	 * Es wird das Netz, welches im Lernverfahren registriert ist,
	 * benutzt.
	 * Am Ende der Methode werden die Observer des NetTrainer benachrichtigt.
	 * 
	 * @param pTrainingAlgorithm Das zu benutzende Lernverfahren und das enthaltene Netz.
	 * @throws PatternDoesNotMatchNetException Falls eines der gegebenen Lerndatenstze
	 * 		nicht der Eingabe-/ Ausgabeschichtgre des Netzes entspricht.
	 */
	public void setTrainingAlgorithm(INetTrainingAlgorithm pTrainingAlgorithm) 
	throws PatternDoesNotMatchNetException{
		if (smLogMethodCall) Log.getInstance().writeln("("+pTrainingAlgorithm.getAlgorithmName()+").", true);
		if (pTrainingAlgorithm != mTrainingAlgorithm){
			mTrainingAlgorithm = pTrainingAlgorithm;
			// check if patterns match net 
			if (mTrainingAlgorithm != null && mTrainingPatterns != null){
				Net lNet = mTrainingAlgorithm.getNet();
				if (lNet.getInputLayerSize() != mTrainingPatterns.getInputPatternSize())
					throw new PatternDoesNotMatchNetException(
						smMSCResourceBundle.getString("INPUTPATTERN_LENGTH")+": "+
						mTrainingPatterns.getInputPatternSize()+" "+
						smMSCResourceBundle.getString("INPUTLAYER_LENGTH")+": "+lNet.getInputLayerSize()
					);
				if (lNet.getOutputLayerSize() != mTrainingPatterns.getOutputPatternSize())
					throw new PatternDoesNotMatchNetException(
						smMSCResourceBundle.getString("OUTPUTPATTERN_LENGTH")+": "+
						mTrainingPatterns.getOutputPatternSize()+" "+
						smMSCResourceBundle.getString("OUTPUTLAYER_LENGTH")+": "+
						lNet.getOutputLayerSize()
					);
			}
			mObserverManager.notifyObservers(this);
		}
	}
	
	/**
	 * Liefert das aktuell benutzte Lernverfahren.
	 * 
	 * @return Das aktuell benutzte Lernverfahren.
	 */
	public INetTrainingAlgorithm getTrainingAlgorithm(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		return mTrainingAlgorithm;
	}
	
	/**
	 * Legt die Lerndatenmenge fest.
	 * Das bergebene Patterns Objekt sollte alle Lerndaten enthalten,
	 * die dem Netz trainiert werden sollen. 
	 * 
	 * @param pTrainingPatterns Die zu lernende Lerndatenstze.
	 * @throws PatternDoesNotMatchNetException Falls eines der gegebenen Lerndatenstze
	 * 		nicht der Eingabe-/ Ausgabeschichtgre des Netzes entspricht.
	 */
	public void setTrainingPatterns(Patterns pTrainingPatterns) 
	throws PatternDoesNotMatchNetException{
		if (smLogMethodCall) Log.getInstance().writeln("("+pTrainingPatterns.getInputPatternSize()+"+"+pTrainingPatterns.getOutputPatternSize()+"x"+pTrainingPatterns.getPatternsCount()+").", true);
		if (pTrainingPatterns != mTrainingPatterns){
			mTrainingPatterns = pTrainingPatterns;
			// check if patterns match net 
			if (mTrainingAlgorithm != null && mTrainingPatterns != null){
				Net lNet = mTrainingAlgorithm.getNet();
				if (lNet.getInputLayerSize() != mTrainingPatterns.getInputPatternSize())
					throw new PatternDoesNotMatchNetException(
						smMSCResourceBundle.getString("INPUTPATTERN_LENGTH")+": "+
						mTrainingPatterns.getInputPatternSize()+" "+
						smMSCResourceBundle.getString("INPUTLAYER_LENGTH")+": "+lNet.getInputLayerSize()
					);
				if (lNet.getOutputLayerSize() != mTrainingPatterns.getOutputPatternSize())
					throw new PatternDoesNotMatchNetException(
						smMSCResourceBundle.getString("OUTPUTPATTERN_LENGTH")+": "+
						mTrainingPatterns.getOutputPatternSize()+" "+
						smMSCResourceBundle.getString("OUTPUTLAYER_LENGTH")+": "+lNet.getOutputLayerSize()
					);
			}
		}
	}
	
	public void setValidationPatterns(Patterns pValidationPatterns) 
	throws PatternDoesNotMatchNetException{
		if (smLogMethodCall) Log.getInstance().writeln("("+pValidationPatterns.getInputPatternSize()+"+"+pValidationPatterns.getOutputPatternSize()+"x"+pValidationPatterns.getPatternsCount()+").", true);
		if (pValidationPatterns != mValidationPatterns){
			mValidationPatterns = pValidationPatterns;
			// check if patterns match net 
			if (mTrainingAlgorithm != null && mValidationPatterns != null){
				Net lNet = mTrainingAlgorithm.getNet();
				if (lNet.getInputLayerSize() != mValidationPatterns.getInputPatternSize())
					throw new PatternDoesNotMatchNetException(
						smMSCResourceBundle.getString("INPUTPATTERN_LENGTH")+": "+
						mValidationPatterns.getInputPatternSize()+" "+
						smMSCResourceBundle.getString("INPUTLAYER_LENGTH")+": "+lNet.getInputLayerSize()
					);
				if (lNet.getOutputLayerSize() != mValidationPatterns.getOutputPatternSize())
					throw new PatternDoesNotMatchNetException(
						smMSCResourceBundle.getString("OUTPUTPATTERN_LENGTH")+": "+
						mValidationPatterns.getOutputPatternSize()+" "+
						smMSCResourceBundle.getString("OUTPUTLAYER_LENGTH")+": "+lNet.getOutputLayerSize()
					);
			}
		}
	}
	
	public Patterns getValidationPatterns() {
		return mValidationPatterns;
	}
	
	/**
	 * Liefert die lernenden Lerndatenstze zurck.
	 * 
	 * @return Die lernenden Lerndatenstze zurck.
	 */
	public Patterns getTrainingPatterns(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		return mTrainingPatterns;
	}
	
	/**
	 * Legt fest, wieviele Lernschritte der Trainer absolvieren soll.
	 * Das nebenlufige Training wird durchgefhrt, bis entweder
	 * die angegebene Anzahl Lernschritte absolviert wurde
	 * oder der durch {@link #setTargetError(double)} definierte Fehlerwert
	 * unterschritten wurde.
	 * 
	 * @param pTargetCycles Anzahl der Lernschritte, die der Trainer absolvieren soll.
	 * @see #setTargetError(double)
	 */
	public void setTargetCycles(int pTargetCycles){
		if (smLogMethodCall) Log.getInstance().writeln("("+pTargetCycles+").", true);
		if (mTargetCycles != pTargetCycles){
			mTargetCycles = pTargetCycles;
			mObserverManager.notifyObservers(this);
		}
	}
	
	/**
	 * Liefert die Anzahl der Lernschritte, die der Trainer absolvieren soll.
	 * 
	 * @return Die Anzahl der Lernschritte, die der Trainer absolvieren soll.
	 */
	public int getTargetCycles(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		return mTargetCycles;
	}
	
	/**
	 * Legt fest, bei welchem Fehlerwert der Trainer das Training als erfolgreich
	 * beenden soll.
	 * Das nebenlufige Training wird durchgefhrt, bis entweder
	 * die durch {@link #setTargetCycles(int)} angegebene Anzahl Lernschritte 
	 * absolviert wurden
	 * oder der definierte Fehlerwert unterschritten wurde.<br/>
	 * Der Typ des Fehlerwertes kann jeder beliebige Hauptfehlertyp
	 * aus der Klasse {@link NetPerformanceStatistics} sein. 
	 * 
	 * @param pTargetError Fehlerwert, dessen Unterschreiten das Training beendet.
	 * @see #setTargetCycles(int)
	 * @see #setTargetErrorType(int)
	 */
	public void setTargetError(double pTargetError){
		if (smLogMethodCall) Log.getInstance().writeln("("+pTargetError+").", true);
		if (mTargetError != pTargetError){
			mTargetError = pTargetError;
			mObserverManager.notifyObservers(this);
		}
	}
	
	/**
	 * Liefert den Fehlerwert, dessen Unterschreiten das Training beendet.
	 * 
	 * @return Fehlerwert, dessen Unterschreiten das Training beendet.
	 */
	public double getTargetError(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		return mTargetError;
	}
	
	/**
	 * Typ des Fehlers, dessen Unterschreitung das Training beendet {@link #setTargetError(double)}.
	 * Als Fehlertyp kann jeder beliebige Hauptfehlertyp
	 * aus der Klasse {@link NetPerformanceStatistics} benutzt werden.
	 * 
	 * @param pTargetErrorType Fehlertyp (ERRORTYPE-Konstante) aus der Klasse {@link NetPerformanceStatistics}
	 * @see #setTargetError(double)
	 * @see #getTargetErrorTypeString()
	 */
	public void setTargetErrorType(int pTargetErrorType){
		if (smLogMethodCall) Log.getInstance().writeln("("+pTargetErrorType+").", true);
		if (pTargetErrorType != mTargetErrorType){
			mTargetErrorType = pTargetErrorType;
			updateLastCalculatedTrainError();
			mObserverManager.notifyObservers(this);
		}
	}
	
	/**
	 * Liefert den Typ des Fehlerwertes aus {@link #setTargetError(double)}.
	 * Der zurckgelieferte Wert entspricht einer Konstante fr den 
	 * Fehlertyp (ERRORTYPE-Konstante) aus der Klasse {@link NetPerformanceStatistics}
	 * 
	 * @return Fehlertyp (ERRORTYPE-Konstante) aus der Klasse {@link NetPerformanceStatistics}
	 * @see #getTargetErrorTypeString()
	 */
	public int getTargetErrorType(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		return mTargetErrorType;
	}
	
	/**
	 * Liefert die Stringreprsentation des aktuellen Fehlertyps zurck.
	 * 
	 * @return Die Stringreprsentation des aktuellen Fehlertyps.
	 */
	public String getTargetErrorTypeString(){
		return NetPerformanceStatistics.STR_errorTypes_short[mTargetErrorType];
	}

	/**
	 * Testet, ob der Trainer zum Trainieren die Lerndatenstze
	 * in gegebener oder in zuflliger Reihenfolge in jedem 
	 * Lernschritt benutzt.
	 * 
	 * @return true, falls eine zufllige Reihenfolge benutzt wird, 
	 * 			false, falls immer die gleiche Reihenfolge benutzt wird.
	 * @see #setUseRandomizedPatternOrder(boolean)
	 */
	public boolean isUsingRandomizedPatternOrder() {
		return mUseRandomizedPatternOrder;
	}

	/**
	 * Legt fest, ob beim Training die Lerndatenstze immer in der gleichen
	 * Reihenfolge gelernt werden sollen, oder ob nach jedem Lernschritt
	 * die Reihenfolge zufllig ermittelt werden soll.<br/>
	 * Eine zufllige Prsentationsreihenfolge der Lerndatenstze erhht
	 * die Generalisierungsfhigkeit des Netzes, was aber geichzeitig bedeutet, 
	 * dass der Netzfehler fr die Lerndatenmenge ebenfalls grer wird.
	 * (Generalisierung vs. Spezialisierung auf die Lerndatenstze).
	 * 
	 * @param pUseRandomizedPatternOrder true = zufllige Reihenfolge benutzen, 
	 * 			false = immer die gleiche Reihenfolge benutzen.
	 * @see #isUsingRandomizedPatternOrder()
	 * @see Patterns#getRandomizedPatternsOrder()
	 */
	public void setUseRandomizedPatternOrder(boolean pUseRandomizedPatternOrder) {
		mUseRandomizedPatternOrder = pUseRandomizedPatternOrder;
	}
	
	/**
	 * Liefert den Fehlerwert aus dem letzten Lernschritt.
	 * Der Typ des Fehlerwertes kann durch die Methode {@link #getTargetErrorType()}
	 * ermittelt werden.
	 * 
	 * @return Aktueller Fehlerwert des Netzes.
	 * @see #getTargetErrorType()
	 * @see #getLastCalcualtedNetStatistics()
	 */
	public double getLastCalculatedError(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		return mLastCalculatedError;	
	}
	
	/**
	 * Liefert die akteullen Fehlerwerte des Netzes. 
	 * 
	 * @return Aktuelle Fehlerwerte des Netzes.
	 */
	public NetPerformanceStatistics getLastCalcualtedNetStatistics(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		return mLastCalculatedTrainNetStatistics;
	}
	
	public boolean hasFinished(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		return (mLastCalculatedTrainNetStatistics != null && (!mRunning || mLastCalculatedError<=mTargetError));
	}
	
	/**
	 * Startet den nebenlufigen Prozess zum Trainiern des Netzes.
	 * Vorher muss dem Trainer ber {@link #setTrainingAlgorithm(INetTrainingAlgorithm)} ein 
	 * Lernverfahren mit Netz und 
	 * ber {@link #setTrainingPatterns(Patterns)} Lerndatenstze zugeteilt werden.
	 * Zustzlich sollte und Aufruf von {@link #setTargetCycles(int)},
	 * {@link #setTargetError(double)}, {@link #setTargetErrorType(int)}
	 * das gewnschte Lernziel festgelegt werden.
	 */
	public void start(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		if (mThread == null) mThread = new Thread(this);
		if (!mRunning){
			mRunning = true;
			mThread.start();
		}
		mObserverManager.notifyObservers(this);
	}
	
	/**
	 * Hlt den nebenlufigen Prozess zum Trainieren des Netzes an. 
	 */
	public void stop(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		mThread = null;
		mRunning = false;
		Log.getInstance().writeln("");
		if (smFinalReport) {
			Log.getInstance().writeln("Net performance:", true);
			Log.getInstance().writeln(NetPerformanceReporter.getNetPerformance(mLastCalculatedTrainNetStatistics, mTrainingAlgorithm.getCycle()));
		}
		mObserverManager.notifyObservers(this);
	}
    
	/**
	 * Prft, ob der nebenlufige Prozess dieses Trainers aktiv ist.
	 * 
	 * @return true, falls dieser Trainer gerade aktiv ist, sonst false.
	 */
	public boolean isRunning(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		return mRunning;
	}
	
	private void updateLastCalculatedTrainError(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		if (mLastCalculatedTrainNetStatistics != null){
			mLastCalculatedTrainError = mLastCalculatedTrainNetStatistics.getErrorByType(mTargetErrorType);
			if (mCheckTargetErrorOnValidationPattern &&  mValidationPatterns!=null) return;
			mLastCalculatedError = mLastCalculatedTrainError;
		}
	}
	
	private void updateLastCalculatedValidationError(){
		if (smLogMethodCall) Log.getInstance().writeln("().", true);
		if (mLastCalculatedValidationNetStatistics != null){
			mLastCalculatedValidationError = mLastCalculatedValidationNetStatistics.getErrorByType(mTargetErrorType);
			if (mValidationPatterns!=null && mCheckTargetErrorOnValidationPattern) {
				mLastCalculatedError = mLastCalculatedValidationError;
			}
			
		}
	}
	
	public void setTrainingStateReport(boolean pDoReport) {
		smLogTrainState = pDoReport;
	}
	
	public void setTrainingStateReportIntervall(int pReportIntervallInCycles) {
		smLogTrainStateIntervall = pReportIntervallInCycles;
	}
	
	public void setFinalReport(boolean pFinalReport) {
		smFinalReport = pFinalReport;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		int lCycle = 1;
		mLastCalculatedTrainError = mTargetError+1D;
		mLastCalculatedError = mTargetError+2D;
		mLastCalculatedValidationError = mTargetError+2D;
		mMinTrainError = mLastCalculatedTrainError+2D;
		mMinValidationError = mLastCalculatedTrainError+2D;
		mMinTrainErrorCycle = 0;
		mMinValidationErrorCycle = 0;
		double [][] lRandomizedInputPatterns = new double[mTrainingPatterns.getPatternsCount()][];
		double [][] lRandomizedOutputPatterns = new double[mTrainingPatterns.getPatternsCount()][];
		int [] lRandomizedOrder;
		int lIndex;
		if (smLogTrainState) Log.getInstance().writeln("Initial stat: "+getTargetErrorTypeString()+": "+mLastCalculatedError+" @ "+mTrainingAlgorithm.getCycle()+" ("+lCycle+"/"+mTargetCycles+")", false);
		do{
			try{
				if (mUseRandomizedPatternOrder){
					lRandomizedOrder = mTrainingPatterns.getRandomizedPatternsOrder();
					for (lIndex = 0; lIndex<lRandomizedOrder.length; lIndex++){
						lRandomizedInputPatterns[lIndex] = mTrainingPatterns.getInputPattern(lRandomizedOrder[lIndex]);
						lRandomizedOutputPatterns[lIndex] = mTrainingPatterns.getOutputPattern(lRandomizedOrder[lIndex]);
					}
					Net.train(
						lRandomizedInputPatterns, 
						lRandomizedOutputPatterns, 
						mTrainingAlgorithm);
				}
				else{
					Net.train(
						mTrainingPatterns.getInputPatterns(), 
						mTrainingPatterns.getOutputPatterns(), 
						mTrainingAlgorithm);
				}
				if (lCycle%mUpdateIntervall==0) {
					checkTrainState(lCycle);
					logTrainState(lCycle);
					if (mLastCalculatedTrainError<mMinTrainError) {
						mMinTrainError = mLastCalculatedTrainError;
						mMinTrainErrorCycle = lCycle;
					}
					if (mLastCalculatedValidationError<mMinValidationError) {
						mMinValidationError = mLastCalculatedValidationError;
						mMinValidationErrorCycle = lCycle;
					}
					mObserverManager.notifyObservers(this);
				}
				lCycle++;
			}
			catch (PatternDoesNotMatchNetException e) {
				// cannot happen, see setTrainingAlgorithm
				e.printStackTrace();
			}
		}
		while (mRunning && mLastCalculatedError>mTargetError && mTargetCycles>=lCycle);
		try{
			checkTrainState(lCycle);
		}
		catch (PatternDoesNotMatchNetException e){
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		logTrainState(lCycle);
		Log.getInstance().writeln("Min Trainingpatterns error: "+mMinTrainError+" @ "+mMinTrainErrorCycle, false);
		if (mValidationPatterns!=null)
			Log.getInstance().writeln("Min Validationpatterns error: "+mMinValidationError+" @ "+mMinValidationErrorCycle, false);
		stop();
	}
	
	private void checkTrainState(int pCycle) throws PatternDoesNotMatchNetException {
		if (mValidationPatterns!=null) {
			mLastCalculatedValidationNetStatistics = NetPerformanceStatisticsCalculator.calculateErrors(
					mTrainingAlgorithm.getNet(), 
					mValidationPatterns.getInputPatterns(), 
					mValidationPatterns.getOutputPatterns());
			updateLastCalculatedValidationError();
		}
		mLastCalculatedTrainNetStatistics = NetPerformanceStatisticsCalculator.calculateErrors(
				mTrainingAlgorithm.getNet(), 
				mTrainingPatterns.getInputPatterns(), 
				mTrainingPatterns.getOutputPatterns());
		updateLastCalculatedTrainError();
	}
	
	private void logTrainState(int pCycle) {
		if (smLogTrainState) {
			if (mValidationPatterns!=null)
				Log.getInstance().writeln(getTargetErrorTypeString()+" Train: "+mLastCalculatedTrainError+"   \tVal: "+mLastCalculatedValidationError+" \t@ "+mTrainingAlgorithm.getCycle()+" ("+pCycle+"/"+mTargetCycles+")", false);
			else Log.getInstance().writeln(getTargetErrorTypeString()+" Train: "+mLastCalculatedTrainError+" \t@ "+mTrainingAlgorithm.getCycle()+" ("+pCycle+"/"+mTargetCycles+")", false);
		}
	}
	
	/* (non-Javadoc)
	 * @see observerPattern.Observable#getObserverManager()
	 */
	public ObserverManager getObserverManager() {
		return mObserverManager;
	}
	
	/**
	 * Trainieren bis die gegebene Anzahl an Lernschritten absolviert wurde.
	 * 
	 * @param pTrainingAlgorithm Lernverfahren, das beim Training bernutzt werden soll.
	 * 		Das Lernverfahren enthlt auch die Referenz zum Netz, das trainiert werden soll. 
	 * @param pInputPatterns Eingabemuster der Lerndatenstze.
	 * @param pOutputPatterns Ausgabemuster der Lerndatenstze (Soll-Werte).
	 * @param pCycles Anzahl der Lernschritte, die absolviert werden sollen.
	 * @throws PatternDoesNotMatchNetException Falls eines der gegebenen Lerndatenstze
	 * 		nicht der Eingabe-/ Ausgabeschichtgre des Netzes entspricht.
	 */
	public static void train(INetTrainingAlgorithm pTrainingAlgorithm,
		double [][] pInputPatterns,	double [][] pOutputPatterns,
		int pCycles) 
		throws PatternDoesNotMatchNetException{
			
		for (int i=0; i<=pCycles; i++)
			Net.train(pInputPatterns, pOutputPatterns, pTrainingAlgorithm);
	}
	
	/**
	 *
	 * Trainieren bis zu einem bestimmten Fehlerwert. Es werden alle Haupt-Fehlertypen
	 * aus {@link mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatistics} 
	 * untersttzt.
	 * 
	 * @param pTrainingAlgorithm Lernverfahren, das beim Training bernutzt werden soll.
	 * 		Das Lernverfahren enthlt auch die Referenz zum Netz, das trainiert werden soll. 
	 * @param pInputPatterns Eingabemuster der Lerndatenstze.
	 * @param pOutputPatterns Ausgabemuster der Lerndatenstze (Soll-Werte).
	 * @param pTargetError Zielwert des Fehlers, bei dem das Training gestoppt werden soll.
	 * @param pErrorType Fehlertyp (ERRORTYPE-Konstante) aus der Klasse {@link NetPerformanceStatistics}
	 * @throws PatternDoesNotMatchNetException Falls eines der gegebenen Lerndatenstze
	 * 		nicht der Eingabe-/ Ausgabeschichtgre des Netzes entspricht.
	 * @see mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatistics
	 */
	public static void train(INetTrainingAlgorithm pTrainingAlgorithm,
		double [][] pInputPatterns,	double [][] pOutputPatterns,
		double pTargetError, int pErrorType)
		throws PatternDoesNotMatchNetException{
		
		NetPerformanceStatistics lNetStat = 
			NetPerformanceStatisticsCalculator.calculateErrors(pTrainingAlgorithm.getNet(), pInputPatterns, pOutputPatterns);
		double lAvgError = lNetStat.getErrorByType(pErrorType);
		while (lAvgError>pTargetError){
			Net.train(pInputPatterns, pOutputPatterns, pTrainingAlgorithm);
			lNetStat = 
				NetPerformanceStatisticsCalculator.calculateErrors(pTrainingAlgorithm.getNet(), pInputPatterns, pOutputPatterns);
			lAvgError = lNetStat.getErrorByType(pErrorType);
		}
	}
	
	/**
	 * Trainieren bis zu einem bestimmten Fehlerwert oder bis eine bestimmte
	 * Anzahl an Lernschritten absolviert wurde. Es werden alle Haupt-Fehlertypen
	 * aus {@link mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatistics} 
	 * untersttzt.<br/>
	 * Je nachdem, welche der beiden Bedingungen, der minimale Fehlerwert oder die
	 * maximale Anzahl der absolvierten Lernschritte, zuerst erreicht wird, beendet
	 * das Training.
	 * 
	 * @param pTrainingAlgorithm Lernverfahren, das beim Training bernutzt werden soll.
	 * 		Das Lernverfahren enthlt auch die Referenz zum Netz, das trainiert werden soll. 
	 * @param pInputPatterns Eingabemuster der Lerndatenstze.
	 * @param pOutputPatterns Ausgabemuster der Lerndatenstze (Soll-Werte).
	 * @param pTargetError Zielwert des Fehlers, bei dem das Training gestoppt werden soll
	 * @param pErrorType Fehlertyp (ERRORTYPE-Konstante) aus der Klasse {@link NetPerformanceStatistics}
	 * @param pCycles Anzahl der Lernschritte, die absolviert werden sollen.
	 * @throws PatternDoesNotMatchNetException Falls eines der gegebenen Lerndatenstze
	 * 		nicht der Eingabe-/ Ausgabeschichtgre des Netzes entspricht.
	 * @see mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatistics
	 */
	public static void train(INetTrainingAlgorithm pTrainingAlgorithm,
		double [][] pInputPatterns,	double [][] pOutputPatterns,
		double pTargetError, int pErrorType, int pCycles)
		throws PatternDoesNotMatchNetException{
		
		// Log.getInstance().writeln("targeterr: "+pTargetError+"  Errtyp: "+pErrorType+" Cyc: "+pCycles, true);
		NetPerformanceStatistics lNetStat = 
			NetPerformanceStatisticsCalculator.calculateErrors(pTrainingAlgorithm.getNet(), pInputPatterns, pOutputPatterns);
		//Log.getInstance().writeln("vor durchlauf", true);
		//Log.getInstance().writeln(NetPerformanceReporter.getNetPerformance(lNetStat, 0));
		
		double lAvgError = lNetStat.getErrorByType(pErrorType);
		//Log.getInstance().writeln("avgerror: "+lAvgError, true);
		int lCycle = 0;
		while (lAvgError>pTargetError && pCycles>=lCycle){
			Net.train(pInputPatterns, pOutputPatterns, pTrainingAlgorithm);
			lNetStat = 
				NetPerformanceStatisticsCalculator.calculateErrors(pTrainingAlgorithm.getNet(), pInputPatterns, pOutputPatterns);
			lAvgError = lNetStat.getErrorByType(pErrorType);
			//System.out.println("errortype: "+pErrorType);
			lCycle++;
			//Log.getInstance().writeln("nach durchlauf", true);
			//System.out.println(NetPerformanceReporter.getNetPerformance(lNetStat, lCycle));
		}
	}
}
