package mscJNeuralNet.trainingAlgorithms;

import java.util.ResourceBundle;

import formats.Formats;

import logger.Log;

/**
 * Das Resilient Backpropagation Lernverfahren nach Riedmiller.<br/>
 * Implentiert nach der Arbeit 
 * <i>Martin Riedmiller, RProp - Description and Implementation Details, 1994</i>
 * <p/>
 * Das Resilient-Propagation Verfahren kombiniert das SuperSAB und das QuickProp 
 * Verfahren, benutzt also eine Art variabler Lernschrittweite (hier der einfachheithalber
 * Alpha ij genannt. In der Literatur und im Kode heisst es Delta ij) 
 * fr jedes einzelne
 * Gewicht und ndert diese nach dem Newtonschen Abstiegsverfahren. Damit
 * ist das RProp Lernverfahren eines der schnellsten bekannten Lernverfahren.
 * <p/>
 * Die Aktualisierung der Kantengewichte wird wie folgt berechnet:<br/>
 * <pre>
 *   Gradientenrichtung = Gradient Wij * Gradient WijAlt
 *   
 *   Falls Gradientenrichtung > 0
 *     Alpha ij = Math.min(Alpha ij * Alpha Plus, Alpha Max)
 * 	 Falls Gradientenrichtung < 0
 *     Alpha ij = Math.max(Alpha ij * Alpha Minus, Alpha Min)
 *   
 *   Falls Gradient Wij > 0
 *     Delta Wij = -1 * Alpha ij
 *   Falls Gradient Wij < 0
 *     Delta Wij = Alpha ij
 *   Falls Gradient Wij = 0
 *     Delta Wij = 0
 * </pre> 
 * <p/>
 * Created on 03.06.2004
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 05.06.2004
 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm
 * @see mscJNeuralNet.trainingAlgorithms.AbstractTrainingAlgorithm
 * @see mscJNeuralNet.trainingAlgorithms
 */
public class RProp extends AbstractTrainingAlgorithm {
	private static final ResourceBundle smMSCResourceBundle = ResourceBundle.getBundle("mscJNeuronalNet_Text");
	private static final ResourceBundle smResourceBundle = ResourceBundle.getBundle("debug");
//	private static final boolean smLogCalculateWeightUpdate;
//	private static final boolean smLogTrain;
	private static final boolean smLogMethodCall;
	static{
//		if (smResourceBundle.getString("showLearningCalculateWeightUpdateRProp").equalsIgnoreCase("true"))
//			smLogCalculateWeightUpdate = true;
//		else smLogCalculateWeightUpdate = false;
		if (smResourceBundle.getString("showMethodCall.TrainingAlgorithm").equalsIgnoreCase("true"))
			smLogMethodCall = true;
		else smLogMethodCall = false;
//		if (smResourceBundle.getString("showTrainRProp").equalsIgnoreCase("true"))
//			smLogTrain = true;
//		else smLogTrain = false;
	}
	
	private double [][] mGradientWOld;
	private double [][] mDeltaIJ;
	private double [][] mWeightChange;
	
	
	private static final String smParameterNames [] = {
		smMSCResourceBundle.getString("DELTA_PLUS"),
		smMSCResourceBundle.getString("DELTA_MINUS"),
		smMSCResourceBundle.getString("DELTA_MAX"),
		smMSCResourceBundle.getString("DELTA_MIN"),
		smMSCResourceBundle.getString("DELTA_ZERO")
	};
	private static final double [] smParameterDefaultValues = {1.2, .5, 50D, .0000001, .1};
	
	private double mDeltaPlus = smParameterDefaultValues[0];
	private double mDeltaMinus = smParameterDefaultValues[1];
	private double mDeltaMax = smParameterDefaultValues[2];
	private double mDeltaMin = smParameterDefaultValues[3];
	private double mDeltaInit = smParameterDefaultValues[4];
	
	/* (non-Javadoc)
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#init()
	 */
	public void init(){
		super.init();
		
		// init gradients
		mGradientWOld = new double [getNet().getAmountNeurons()][getNet().getAmountNeurons()];
		mDeltaIJ = new double [getNet().getAmountNeurons()][getNet().getAmountNeurons()];
		mWeightChange = new double [getNet().getAmountNeurons()][getNet().getAmountNeurons()];
		for (int x=0; x<getNet().getAmountNeurons(); x++){
			for (int y=0; y<getNet().getAmountNeurons(); y++){
				mGradientWOld[x][y] = 0D;
				mDeltaIJ[x][y] = mDeltaInit;
				mWeightChange[x][y] = 0D;
			}
		}
		setCycle(0);
	}

	private double mGradientDirection;

	/* (non-Javadoc)
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#applyWeightUpdate(int, int)
	 */
	public void applyWeightUpdate(int pOutputting, int pInputting){
		super.applyWeightUpdate(pOutputting, pInputting);
		
		mGradientDirection = getActualGradients()[pOutputting][pInputting] * 
			mGradientWOld[pOutputting][pInputting];
		if (mGradientDirection > 0)
			mDeltaIJ[pOutputting][pInputting] = 
				Math.min(
					mDeltaIJ[pOutputting][pInputting]*mDeltaPlus,
					mDeltaMax);
		else if (mGradientDirection < 0)
			mDeltaIJ[pOutputting][pInputting] =
				Math.max(
					mDeltaIJ[pOutputting][pInputting]*mDeltaMinus,
					mDeltaMin);
		
		if (getActualGradients()[pOutputting][pInputting]>0)
			mWeightChange[pOutputting][pInputting] = -1D*mDeltaIJ[pOutputting][pInputting];
		else if (getActualGradients()[pOutputting][pInputting]<0)
			mWeightChange[pOutputting][pInputting] = mDeltaIJ[pOutputting][pInputting];
		else mWeightChange[pOutputting][pInputting] = 0D;
		
		getNet().getWeights()[pOutputting][pInputting] += mWeightChange[pOutputting][pInputting];
		mGradientWOld[pOutputting][pInputting] = getActualGradients()[pOutputting][pInputting];
	}
	
	/**
	 * Dieses Lernverfahren untersttzt folgende Parameter: <br/>
	 * <pre>
	 * Index  Bezeichnung   Funktion
	 *   0    Delta Plus    Faktor zum Erhhen der variablen Lernschrittweite
	 *   1    Delta Minus   Faktor zum Verkleinern der variablen Lernschrittweite
	 *   2    Delta Max     Maximale variable Lernschrittweite
	 *   3    Delta Min     Minimale variable Lernschrittweite
	 *   4    Initial Delta Initialwert fr alle variablen Lernschrittweiten
	 * </pre>
	 * Am Ende der Methode werden die Observer benachrichtigt.
	 *  
	 * @see INetTrainingAlgorithm#setLearningParameters(double[])
	 */
	public void setLearningParameters(double[] pLearningParameters) {
		if (smLogMethodCall) Log.getInstance().writeln(this.getClass().getName()+".setLearningParameters("+Formats.format(pLearningParameters)+")");
		Log.getInstance().writeln(this.getClass().getName()+".setLearningParameters("+Formats.format(pLearningParameters)+")");
		if (pLearningParameters.length>0){
			if (pLearningParameters[0] >= 0D) mDeltaPlus = pLearningParameters[0];
			if (pLearningParameters.length>1){
				if (pLearningParameters[1] >= 0D) mDeltaMinus = pLearningParameters[1];
				if (pLearningParameters.length>2){
					if (pLearningParameters[2] >= 0D) mDeltaMax = pLearningParameters[2];
					if (pLearningParameters.length>3){
						if (pLearningParameters[3] >= 0D) mDeltaMin = pLearningParameters[3];
						if (pLearningParameters.length>4){
							if (pLearningParameters[4] >= 0D) mDeltaInit = pLearningParameters[4];
						}
					}
				}
			}
		}
		getObserverManager().notifyObservers(this);
	}
	
	/**
	 * Dieses Lernverfahren untersttzt folgende Parameter: <br/>
	 * <pre>
	 * Index  Bezeichnung   Funktion
	 *   0    Delta Plus    Faktor zum Erhhen der variablen Lernschrittweite
	 *   1    Delta Minus   Faktor zum Verkleinern der variablen Lernschrittweite
	 *   2    Delta Max     Maximale variable Lernschrittweite
	 *   3    Delta Min     Minimale variable Lernschrittweite
	 *   4    Initial Delta Initialwert fr alle variablen Lernschrittweiten
	 * </pre>
	 * 
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#getLearningParameters()
	 */
	public double[] getLearningParameters() {
		double [] lParameters = new double[smParameterNames.length];
		lParameters[0] = mDeltaPlus;
		lParameters[1] = mDeltaMinus;
		lParameters[2] = mDeltaMax;
		lParameters[3] = mDeltaMin;
		lParameters[4] = mDeltaInit;
		return lParameters;
	}

	/**
	 * Dieses Lernverfahren untersttzt folgende Parameter: <br/>
	 * <pre>
	 * Index  Bezeichnung   Funktion
	 *   0    Delta Plus    Faktor zum Erhhen der variablen Lernschrittweite
	 *   1    Delta Minus   Faktor zum Verkleinern der variablen Lernschrittweite
	 *   2    Delta Max     Maximale variable Lernschrittweite
	 *   3    Delta Min     Minimale variable Lernschrittweite
	 *   4    Initial Delta Initialwert fr alle variablen Lernschrittweiten
	 * </pre>
	 * 
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#getLearningParameterNames()
	 */
	public String[] getLearningParameterNames() {
		return smParameterNames;
	}
	
	/**
	 * Dieses Lernverfahren untersttzt folgende Parameter: <br/>
	 * <pre>
	 * Index  Bezeichnung   Funktion
	 *   0    Delta Plus    Faktor zum Erhhen der variablen Lernschrittweite
	 *   1    Delta Minus   Faktor zum Verkleinern der variablen Lernschrittweite
	 *   2    Delta Max     Maximale variable Lernschrittweite
	 *   3    Delta Min     Minimale variable Lernschrittweite
	 *   4    Initial Delta Initialwert fr alle variablen Lernschrittweiten
	 * </pre>
	 * 
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#getDefaultLearningParameters()
	 */
	public double [] getDefaultLearningParameters(){
		return smParameterDefaultValues;
	}
	
	/**
	 * Liefert den String "RProp" zurck.
	 * 
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#getAlgorithmName()
	 */
	public String getAlgorithmName(){
		return "RProp";
	}
	
	public String toString(){
		return "RProp d+: "+mDeltaPlus+" d-: "+mDeltaMinus+" d0: "+mDeltaInit+" dmax: "+mDeltaMax+" dmin: "+mDeltaMin ;
	}
}
